@@ -23,7 +23,7 @@
  * @title: McpAccountStorage
  * @short_description: Account Storage object, implemented by plugins
  * @see_also:
- * @include: mission-control-plugins/account-storage.h
+ * @include: mission-control-plugins/mission-control-plugins.h
  *
  * Plugins may implement #McpAccountStorage in order to provide account
  * parameter storage backends to the AccountManager object.
@@ -78,17 +78,20 @@
 #include <mission-control-plugins/mission-control-plugins.h>
 #include <mission-control-plugins/mcp-signals-marshal.h>
 #include <mission-control-plugins/implementation.h>
+#include <mission-control-plugins/debug-internal.h>
 #include <glib.h>
 
+#define MCP_DEBUG_TYPE  MCP_DEBUG_ACCOUNT_STORAGE
+
 #ifdef ENABLE_DEBUG
 
-#define DEBUG(_p, _format, ...) \
-  g_debug ("%s: %s: " _format, G_STRFUNC, \
-      (_p != NULL) ? mcp_account_storage_name (_p) : "NULL", ##__VA_ARGS__)
+#define SDEBUG(_p, _format, ...) \
+  DEBUG("%s: " _format, \
+        (_p != NULL) ? mcp_account_storage_name (_p) : "NULL", ##__VA_ARGS__)
 
 #else  /* ENABLE_DEBUG */
 
-#define DEBUG(_p, _format, ...) do {} while (0);
+#define SDEBUG(_p, _format, ...) do {} while (0);
 
 #endif /* ENABLE_DEBUG */
 
@@ -104,65 +107,6 @@ enum
 
 static guint signals[NO_SIGNAL] = { 0 };
 
-struct _McpAccountStorageIface
-{
-  GTypeInterface parent;
-
-  gint priority;
-  const gchar *name;
-  const gchar *desc;
-  const gchar *provider;
-
-  gboolean (*set) (
-      const McpAccountStorage *self,
-      const McpAccountManager *am,
-      const gchar *account,
-      const gchar *key,
-      const gchar *val);
-
-  gboolean (*get) (
-      const McpAccountStorage *self,
-      const McpAccountManager *am,
-      const gchar *account,
-      const gchar *key);
-
-  gboolean (*delete) (
-      const McpAccountStorage *self,
-      const McpAccountManager *am,
-      const gchar *account,
-      const gchar *key);
-
-  gboolean (*commit) (
-      const McpAccountStorage *self,
-      const McpAccountManager *am);
-
-  GList * (*list) (
-      const McpAccountStorage *self,
-      const McpAccountManager *am);
-
-  void (*ready) (
-      const McpAccountStorage *self,
-      const McpAccountManager *am);
-
-  gboolean (*commit_one) (
-      const McpAccountStorage *self,
-      const McpAccountManager *am,
-      const gchar *account);
-
-  void (*get_identifier) (
-      const McpAccountStorage *self,
-      const gchar *account,
-      GValue *identifier);
-
-  GHashTable *(*get_additional_info) (
-      const McpAccountStorage *self,
-      const gchar *account);
-
-  TpStorageRestrictionFlags (*get_restrictions) (
-      const McpAccountStorage *self,
-      const gchar *account);
-};
-
 static void
 class_init (gpointer klass,
     gpointer data)
@@ -315,71 +259,49 @@ mcp_account_storage_iface_set_provider (McpAccountStorageIface *iface,
 
 void
 mcp_account_storage_iface_implement_get (McpAccountStorageIface *iface,
-    gboolean (*method) (
-        const McpAccountStorage *,
-        const McpAccountManager *,
-        const gchar *,
-        const gchar *))
+    McpAccountStorageGetFunc method)
 {
   iface->get = method;
 }
 
 void
 mcp_account_storage_iface_implement_set (McpAccountStorageIface *iface,
-    gboolean (*method) (
-        const McpAccountStorage *,
-        const McpAccountManager *,
-        const gchar *,
-        const gchar *,
-        const gchar *))
+    McpAccountStorageSetFunc method)
 {
   iface->set = method;
 }
 
 void
 mcp_account_storage_iface_implement_delete (McpAccountStorageIface *iface,
-    gboolean (*method) (
-        const McpAccountStorage *,
-        const McpAccountManager *,
-        const gchar *,
-        const gchar *))
+    McpAccountStorageDeleteFunc method)
 {
   iface->delete = method;
 }
 
 void
 mcp_account_storage_iface_implement_commit (McpAccountStorageIface *iface,
-    gboolean (*method) (
-        const McpAccountStorage *,
-        const McpAccountManager *))
+    McpAccountStorageCommitFunc method)
 {
   iface->commit = method;
 }
 
 void
 mcp_account_storage_iface_implement_commit_one (McpAccountStorageIface *iface,
-    gboolean (*method) (
-        const McpAccountStorage *,
-        const McpAccountManager *,
-        const gchar *))
+    McpAccountStorageCommitOneFunc method)
 {
   iface->commit_one = method;
 }
 
 void
 mcp_account_storage_iface_implement_list (McpAccountStorageIface *iface,
-    GList * (*method) (
-        const McpAccountStorage *,
-        const McpAccountManager *))
+    McpAccountStorageListFunc method)
 {
   iface->list = method;
 }
 
 void
 mcp_account_storage_iface_implement_ready (McpAccountStorageIface *iface,
-    void  (*method) (
-        const McpAccountStorage *,
-        const McpAccountManager *))
+    McpAccountStorageReadyFunc method)
 {
   iface->ready = method;
 }
@@ -387,10 +309,7 @@ mcp_account_storage_iface_implement_ready (McpAccountStorageIface *iface,
 void
 mcp_account_storage_iface_implement_get_identifier (
     McpAccountStorageIface *iface,
-    void (*method) (
-        const McpAccountStorage *,
-        const gchar *,
-        GValue *))
+    McpAccountStorageGetIdentifierFunc method)
 {
   iface->get_identifier = method;
 }
@@ -398,9 +317,7 @@ mcp_account_storage_iface_implement_get_identifier (
 void
 mcp_account_storage_iface_implement_get_additional_info (
     McpAccountStorageIface *iface,
-    GHashTable *(*method) (
-        const McpAccountStorage *,
-        const gchar *))
+    McpAccountStorageGetAdditionalInfoFunc method)
 {
   iface->get_additional_info = method;
 }
@@ -408,9 +325,7 @@ mcp_account_storage_iface_implement_get_additional_info (
 void
 mcp_account_storage_iface_implement_get_restrictions (
     McpAccountStorageIface *iface,
-    guint (*method) (
-        const McpAccountStorage *,
-        const gchar *))
+    McpAccountStorageGetRestrictionsFunc method)
 {
   iface->get_restrictions = method;
 }
@@ -484,7 +399,7 @@ mcp_account_storage_get (const McpAccountStorage *storage,
 {
   McpAccountStorageIface *iface = MCP_ACCOUNT_STORAGE_GET_IFACE (storage);
 
-  DEBUG (storage, "");
+  SDEBUG (storage, "");
   g_return_val_if_fail (iface != NULL, FALSE);
 
   return iface->get (storage, am, account, key);
@@ -516,7 +431,7 @@ mcp_account_storage_set (const McpAccountStorage *storage,
 {
   McpAccountStorageIface *iface = MCP_ACCOUNT_STORAGE_GET_IFACE (storage);
 
-  DEBUG (storage, "");
+  SDEBUG (storage, "");
   g_return_val_if_fail (iface != NULL, FALSE);
 
   return iface->set (storage, am, account, key, val);
@@ -553,7 +468,7 @@ mcp_account_storage_delete (const McpAccountStorage *storage,
 {
   McpAccountStorageIface *iface = MCP_ACCOUNT_STORAGE_GET_IFACE (storage);
 
-  DEBUG (storage, "");
+  SDEBUG (storage, "");
   g_return_val_if_fail (iface != NULL, FALSE);
 
   return iface->delete (storage, am, account, key);
@@ -585,16 +500,23 @@ mcp_account_storage_commit (const McpAccountStorage *storage,
 {
   McpAccountStorageIface *iface = MCP_ACCOUNT_STORAGE_GET_IFACE (storage);
 
-  DEBUG (storage, "");
+  SDEBUG (storage, "committing all accounts");
   g_return_val_if_fail (iface != NULL, FALSE);
 
-  if (iface->commit == NULL)
+  if (iface->commit != NULL)
     {
-      DEBUG (storage, "no commit method implemented, cannot save accounts");
+      return iface->commit (storage, am);
+    }
+  else if (iface->commit_one != NULL)
+    {
+      return iface->commit_one (storage, am, NULL);
+    }
+  else
+    {
+      SDEBUG (storage,
+          "neither commit nor commit_one is implemented; cannot save accounts");
       return FALSE;
     }
-
-  return iface->commit (storage, am);
 }
 
 /**
@@ -618,16 +540,14 @@ mcp_account_storage_commit_one (const McpAccountStorage *storage,
 {
   McpAccountStorageIface *iface = MCP_ACCOUNT_STORAGE_GET_IFACE (storage);
 
-  DEBUG (storage, "");
+  SDEBUG (storage, "called for %s", account ? account : "<all accounts>");
   g_return_val_if_fail (iface != NULL, FALSE);
 
-  /* if the plugin doesn't implement commit_one _or_ we weren't asked to  *
-   * save a specific account, try to commit them all: if that can't work, *
-   * give up, as there's nothing else that can be done:                   */
-  if (iface->commit_one == NULL || account == NULL)
+  if (iface->commit_one != NULL)
+    return iface->commit_one (storage, am, account);
+  else
+    /* Fall back to plain ->commit() */
     return mcp_account_storage_commit (storage, am);
-
-  return iface->commit_one (storage, am, account);
 }
 
 /**
@@ -648,7 +568,7 @@ mcp_account_storage_list (const McpAccountStorage *storage,
 {
   McpAccountStorageIface *iface = MCP_ACCOUNT_STORAGE_GET_IFACE (storage);
 
-  DEBUG (storage, "");
+  SDEBUG (storage, "");
   g_return_val_if_fail (iface != NULL, NULL);
 
   return iface->list (storage, am);
@@ -694,7 +614,7 @@ mcp_account_storage_get_identifier (const McpAccountStorage *storage,
 {
   McpAccountStorageIface *iface = MCP_ACCOUNT_STORAGE_GET_IFACE (storage);
 
-  DEBUG (storage, "");
+  SDEBUG (storage, "");
   g_return_if_fail (iface != NULL);
   g_return_if_fail (identifier != NULL);
   g_return_if_fail (!G_IS_VALUE (identifier));
@@ -728,7 +648,7 @@ mcp_account_storage_get_additional_info (const McpAccountStorage *storage,
   McpAccountStorageIface *iface = MCP_ACCOUNT_STORAGE_GET_IFACE (storage);
   GHashTable *ret = NULL;
 
-  DEBUG (storage, "");
+  SDEBUG (storage, "");
   g_return_val_if_fail (iface != NULL, FALSE);
 
   if (iface->get_additional_info != NULL)
@@ -745,10 +665,11 @@ mcp_account_storage_get_additional_info (const McpAccountStorage *storage,
  * @storage: an #McpAccountStorage instance
  * @account: the unique name of the account
  *
- * Returns: a bitmask with the restrictions of this
- * account storage.
+ * Returns: a bitmask of %TpStorageRestrictionFlags with the restrictions to
+ *  account storage.
  */
-TpStorageRestrictionFlags
+/* FIXME: when breaking API, make this return TpStorageRestrictionFlags */
+guint
 mcp_account_storage_get_restrictions (const McpAccountStorage *storage,
     const gchar *account)
 {
