@@ -1,7 +1,7 @@
 /* Map containing registered Telepathy clients
  *
- * Copyright (C) 2007-2009 Nokia Corporation.
- * Copyright (C) 2009 Collabora Ltd.
+ * Copyright © 2007-2011 Nokia Corporation.
+ * Copyright © 2009-2011 Collabora Ltd.
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public License
@@ -21,11 +21,15 @@
 
 #include "client-registry.h"
 
-#include <telepathy-glib/defs.h>
 #include <telepathy-glib/handle-repo-dynamic.h>
+#include <telepathy-glib/telepathy-glib.h>
 
 #include "mcd-debug.h"
 
+#include <dbus/dbus.h>
+#include <dbus/dbus-glib.h>
+#include <dbus/dbus-glib-lowlevel.h>
+
 G_DEFINE_TYPE (McdClientRegistry, _mcd_client_registry, G_TYPE_OBJECT)
 
 enum
@@ -301,25 +305,84 @@ mcd_client_registry_list_names_cb (TpDBusDaemon *proxy,
    * ReloadConfig), so simplify by doing nothing */
 }
 
-static void
-mcd_client_registry_name_owner_changed_cb (TpDBusDaemon *proxy,
-    const gchar *name,
-    const gchar *old_owner,
-    const gchar *new_owner,
-    gpointer user_data G_GNUC_UNUSED,
-    GObject *weak_object)
+static DBusHandlerResult
+mcd_client_registry_name_owner_filter (DBusConnection *conn,
+    DBusMessage *msg,
+    gpointer data)
 {
-  McdClientRegistry *self = MCD_CLIENT_REGISTRY (weak_object);
+  McdClientRegistry *self = MCD_CLIENT_REGISTRY (data);
 
-  /* dbus-glib guarantees this */
-  g_assert (name != NULL);
-  g_assert (old_owner != NULL);
-  g_assert (new_owner != NULL);
-
-  if (old_owner[0] == '\0' && new_owner[0] != '\0')
+  /* make sure this is the right kind of signal: */
+  if (dbus_message_is_signal (msg, DBUS_INTERFACE_DBUS, "NameOwnerChanged"))
     {
-      _mcd_client_registry_found_name (self, name, new_owner, FALSE);
+      const gchar *dbus_name = NULL;
+      const gchar *old_owner = NULL;
+      const gchar *new_owner = NULL;
+      gboolean ok = dbus_message_get_args (msg, NULL,
+          DBUS_TYPE_STRING, &dbus_name,
+          DBUS_TYPE_STRING, &old_owner,
+          DBUS_TYPE_STRING, &new_owner,
+          DBUS_TYPE_INVALID);
+
+      /* could not unpack args -> invalid -> stop processing right here */
+      if (!ok)
+        return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
+
+      if (tp_str_empty (old_owner) && !tp_str_empty (new_owner))
+        _mcd_client_registry_found_name (self, dbus_name, new_owner, FALSE);
     }
+
+  /* in case somebody else is also interested */
+  return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
+}
+
+static gboolean
+add_match (DBusConnection *conn,
+    const gchar const *rule,
+    const gchar *msg)
+{
+  DBusError error = { 0 };
+
+  dbus_error_init (&error);
+  dbus_bus_add_match (conn, rule, &error);
+
+  if (dbus_error_is_set (&error))
+    {
+      g_warning ("Could not add %s match rule: %s", msg, error.message);
+      dbus_error_free (&error);
+      return FALSE;
+    }
+
+  return TRUE;
+}
+
+static void
+watch_clients (McdClientRegistry *self)
+{
+  TpDBusDaemon *dbus_daemon = self->priv->dbus_daemon;
+  DBusGConnection *gconn = tp_proxy_get_dbus_connection (dbus_daemon);
+  DBusConnection *dconn = dbus_g_connection_get_connection (gconn);
+  gboolean arg0_filtered = FALSE;
+
+#define MATCH_ITEM(t,x) #t "='" x "'"
+
+#define NAME_OWNER_RULE \
+    MATCH_ITEM (type,      "signal")            "," \
+    MATCH_ITEM (sender,    DBUS_SERVICE_DBUS)   "," \
+    MATCH_ITEM (interface, DBUS_INTERFACE_DBUS) "," \
+    MATCH_ITEM (member,    "NameOwnerChanged")
+
+#define CLIENT_MATCH_RULE \
+    NAME_OWNER_RULE "," \
+    MATCH_ITEM (arg0namespace, "org.freedesktop.Telepathy.Client")
+
+  arg0_filtered = dbus_connection_add_filter (dconn,
+      mcd_client_registry_name_owner_filter,
+      self,
+      NULL);
+
+  if (arg0_filtered && !add_match (dconn, CLIENT_MATCH_RULE, "client names"))
+    add_match (dconn, NAME_OWNER_RULE, "all dbus names");
 }
 
 static void
@@ -336,11 +399,7 @@ mcd_client_registry_constructed (GObject *object)
 
   DEBUG ("Starting to look for clients");
 
-  /* FIXME: ideally, this would be a more specific match, using arg0prefix
-   * (when dbus-daemon supports that, which it doesn't yet) so we only get
-   * new clients. */
-  tp_cli_dbus_daemon_connect_to_name_owner_changed (self->priv->dbus_daemon,
-      mcd_client_registry_name_owner_changed_cb, NULL, NULL, object, NULL);
+  watch_clients (self);
 
   tp_cli_dbus_daemon_call_list_names (self->priv->dbus_daemon, -1,
       mcd_client_registry_list_names_cb, NULL, NULL, object);
@@ -400,25 +459,27 @@ mcd_client_registry_dispose (GObject *object)
 
   if (self->priv->dbus_daemon != NULL)
     {
-      g_object_unref (self->priv->dbus_daemon);
-      self->priv->dbus_daemon = NULL;
+      DBusGConnection *gconn =
+        tp_proxy_get_dbus_connection (self->priv->dbus_daemon);
+      DBusConnection *dconn = dbus_g_connection_get_connection (gconn);
+
+      dbus_connection_remove_filter (dconn,
+          mcd_client_registry_name_owner_filter,
+          self);
     }
 
-  if (self->priv->string_pool != NULL)
-    {
-      g_object_unref (self->priv->string_pool);
-      self->priv->string_pool = NULL;
-    }
+  tp_clear_object (&self->priv->dbus_daemon);
+  tp_clear_object (&self->priv->string_pool);
 
   if (self->priv->clients != NULL)
     {
       g_hash_table_foreach (self->priv->clients,
           mcd_client_registry_disconnect_client_signals, self);
 
-      g_hash_table_destroy (self->priv->clients);
-      self->priv->clients = NULL;
     }
 
+  tp_clear_pointer (&self->priv->clients, g_hash_table_destroy);
+
   if (chain_up != NULL)
     chain_up (object);
 }
@@ -511,3 +572,177 @@ _mcd_client_registry_is_ready (McdClientRegistry *self)
   g_return_val_if_fail (MCD_IS_CLIENT_REGISTRY (self), FALSE);
   return self->priv->startup_completed;
 }
+
+typedef struct
+{
+    McdClientProxy *client;
+    gboolean bypass;
+    gsize quality;
+} PossibleHandler;
+
+static gint
+possible_handler_cmp (gconstpointer a_,
+                      gconstpointer b_)
+{
+  const PossibleHandler *a = a_;
+  const PossibleHandler *b = b_;
+
+  if (a->bypass)
+    {
+      if (!b->bypass)
+        {
+          /* BypassApproval wins, so a is better than b */
+          return 1;
+        }
+    }
+  else if (b->bypass)
+    {
+      /* BypassApproval wins, so b is better than a */
+      return -1;
+    }
+
+  if (a->quality < b->quality)
+    {
+      return -1;
+    }
+
+  if (b->quality < a->quality)
+    {
+      return 1;
+    }
+
+  return 0;
+}
+
+GList *
+_mcd_client_registry_list_possible_handlers (McdClientRegistry *self,
+    const gchar *preferred_handler,
+    GHashTable *request_props,
+    const GList *channels,
+    const gchar *must_have_unique_name)
+{
+  GList *handlers = NULL;
+  const GList *iter;
+  GList *handlers_iter;
+  GHashTableIter client_iter;
+  gpointer client_p;
+
+  _mcd_client_registry_init_hash_iter (self, &client_iter);
+
+  while (g_hash_table_iter_next (&client_iter, NULL, &client_p))
+    {
+      McdClientProxy *client = MCD_CLIENT_PROXY (client_p);
+      gsize total_quality = 0;
+
+      if (must_have_unique_name != NULL &&
+          tp_strdiff (must_have_unique_name,
+            _mcd_client_proxy_get_unique_name (client)))
+        {
+          /* we're trying to redispatch to an existing handler, and this is
+           * not it */
+          continue;
+        }
+
+      if (!tp_proxy_has_interface_by_id (client,
+            TP_IFACE_QUARK_CLIENT_HANDLER))
+        {
+            /* not a handler at all */
+            continue;
+        }
+
+      if (channels == NULL)
+        {
+          /* We don't know any channels' properties (the next loop will not
+           * execute), so we must work out the quality of match from the
+           * channel request. We can assume that the request will return one
+           * channel, with the requested properties, plus Requested == TRUE.
+           */
+          g_assert (request_props != NULL);
+          total_quality = _mcd_client_match_filters (request_props,
+              _mcd_client_proxy_get_handler_filters (client), TRUE);
+        }
+
+      for (iter = channels; iter != NULL; iter = iter->next)
+        {
+          TpChannel *channel = iter->data;
+          GHashTable *properties;
+          guint quality;
+
+          g_assert (TP_IS_CHANNEL (channel));
+          properties = tp_channel_borrow_immutable_properties (channel);
+
+          quality = _mcd_client_match_filters (properties,
+              _mcd_client_proxy_get_handler_filters (client), FALSE);
+
+          if (quality == 0)
+            {
+              total_quality = 0;
+              break;
+            }
+          else
+            {
+              total_quality += quality;
+            }
+        }
+
+      if (total_quality > 0)
+        {
+          PossibleHandler *ph = g_slice_new0 (PossibleHandler);
+
+          ph->client = client;
+          ph->bypass = _mcd_client_proxy_get_bypass_approval (client);
+          ph->quality = total_quality;
+
+          handlers = g_list_prepend (handlers, ph);
+        }
+    }
+
+  /* if no handlers can take them all, fail - unless we're operating on
+   * a request that specified a preferred handler, in which case assume
+   * it's suitable */
+  if (handlers == NULL)
+    {
+      McdClientProxy *client;
+
+      if (preferred_handler == NULL || preferred_handler[0] == '\0')
+        {
+          return NULL;
+        }
+
+      client = _mcd_client_registry_lookup (self, preferred_handler);
+
+      if (client == NULL)
+        {
+          return NULL;
+        }
+
+      return g_list_append (NULL, client);
+    }
+
+  /* We have at least one handler that can take the whole batch. Sort
+   * the possible handlers, most preferred first (i.e. sort by ascending
+   * quality then reverse) */
+  handlers = g_list_sort (handlers, possible_handler_cmp);
+  handlers = g_list_reverse (handlers);
+
+  /* convert in-place from a list of PossibleHandler to a list of
+   * McdClientProxy */
+  for (handlers_iter = handlers;
+       handlers_iter != NULL;
+       handlers_iter = handlers_iter->next)
+    {
+      PossibleHandler *ph = handlers_iter->data;
+
+      handlers_iter->data = ph->client;
+      g_slice_free (PossibleHandler, ph);
+    }
+
+  return handlers;
+}
+
+TpDBusDaemon *
+_mcd_client_registry_get_dbus_daemon (McdClientRegistry *self)
+{
+    g_return_val_if_fail (MCD_IS_CLIENT_REGISTRY (self), NULL);
+    return self->priv->dbus_daemon;
+}
