@@ -434,6 +434,7 @@ static void _sso_updated (AgAccount *account,
   const gchar *deleted_key;
   guint i;
   gboolean params_updated = FALSE;
+  const gchar *immutables[] = { MC_SERVICE_KEY, SERVICES_KEY, NULL };
 
   /* account has no name yet: might be time to create it */
   if (name == NULL)
@@ -496,6 +497,15 @@ static void _sso_updated (AgAccount *account,
         }
     }
 
+  /* special case values always exist and therefore cannot be deleted */
+  for (i = 0; immutables[i] != NULL; i++)
+    {
+      Setting *immutable = setting_data (immutables[i], SETTING_AG);
+
+      g_hash_table_remove (unseen, immutable->ag_name);
+      clear_setting_data (immutable);
+    }
+
   /* signal (and update) deleted settings: */
   g_hash_table_iter_init (&deleted_iter, unseen);
 
@@ -509,6 +519,9 @@ static void _sso_updated (AgAccount *account,
         g_signal_emit_by_name (mcpa, "altered-one", name, deleted_key);
     }
 
+  g_hash_table_unref (unseen);
+  g_strfreev (keys);
+
   if (params_updated)
     g_signal_emit_by_name (mcpa, "altered-one", name, "Parameters");
 
