@@ -3,8 +3,8 @@
 /*
  * Mission Control client proxy.
  *
- * Copyright (C) 2009 Nokia Corporation
- * Copyright (C) 2009 Collabora Ltd.
+ * Copyright (C) 2009-2010 Nokia Corporation
+ * Copyright (C) 2009-2010 Collabora Ltd.
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -76,6 +76,8 @@ struct _McdClientProxyPrivate
     gboolean introspect_started;
     gboolean ready;
     gboolean bypass_approval;
+    gboolean bypass_observers;
+    gboolean delay_approvers;
     gboolean recover;
 
     /* If a client was in the ListActivatableNames list, it must not be
@@ -346,8 +348,6 @@ parse_client_file (McdClientProxy *client,
     GList *approver_filters = NULL;
     GList *observer_filters = NULL;
     GList *handler_filters = NULL;
-    gboolean bypass;
-    gboolean recover;
 
     iface_names = g_key_file_get_string_list (file, TP_IFACE_CLIENT,
                                               "Interfaces", 0, NULL);
@@ -404,13 +404,21 @@ parse_client_file (McdClientProxy *client,
                                             handler_filters);
 
     /* Other client options */
-    bypass = g_key_file_get_boolean (file, TP_IFACE_CLIENT_HANDLER,
-                                     "BypassApproval", NULL);
-    client->priv->bypass_approval = bypass;
+    client->priv->bypass_approval =
+        g_key_file_get_boolean (file, TP_IFACE_CLIENT_HANDLER,
+                                "BypassApproval", NULL);
 
-    recover = g_key_file_get_boolean (file, TP_IFACE_CLIENT_OBSERVER,
-                                      "Recover", NULL);
-    client->priv->recover = recover;
+    client->priv->bypass_observers =
+        g_key_file_get_boolean (file, TP_IFACE_CLIENT_HANDLER,
+                                "BypassObservers", NULL);
+
+    client->priv->delay_approvers =
+        g_key_file_get_boolean (file, TP_IFACE_CLIENT_OBSERVER,
+                                "DelayApprovers", NULL);
+
+    client->priv->recover =
+        g_key_file_get_boolean (file, TP_IFACE_CLIENT_OBSERVER,
+                                "Recover", NULL);
 
     cap_tokens = g_key_file_get_keys (file,
                                       TP_IFACE_CLIENT_HANDLER ".Capabilities",
@@ -678,6 +686,10 @@ _mcd_client_proxy_handler_get_all_cb (TpProxy *proxy,
     self->priv->bypass_approval = bypass;
     DEBUG ("%s has BypassApproval=%c", bus_name, bypass ? 'T' : 'F');
 
+    bypass = tp_asv_get_boolean (properties, "BypassObservers", NULL);
+    self->priv->bypass_observers = bypass;
+    DEBUG ("%s has BypassObservers=%c", bus_name, bypass ? 'T' : 'F');
+
     /* don't emit handler-capabilities-changed if we're not actually available
      * any more - if that's the case, then we already signalled our loss of
      * any capabilities */
@@ -770,6 +782,12 @@ _mcd_client_proxy_observer_get_all_cb (TpProxy *proxy,
     /* by now, we at least know whether the client is running or not */
     g_assert (self->priv->unique_name != NULL);
 
+    /* FALSE if DelayApprovers is invalid or missing is a good fallback */
+    self->priv->delay_approvers = tp_asv_get_boolean (
+        properties, "DelayApprovers", NULL);
+    DEBUG ("%s has DelayApprovers=%c", bus_name,
+        self->priv->delay_approvers ? 'T' : 'F');
+
     filters = tp_asv_get_boxed (properties, "ObserverChannelFilter",
                                 TP_ARRAY_TYPE_STRING_VARIANT_MAP_LIST);
 
@@ -1396,6 +1414,22 @@ _mcd_client_proxy_get_bypass_approval (McdClientProxy *self)
     return self->priv->bypass_approval;
 }
 
+gboolean
+_mcd_client_proxy_get_bypass_observers (McdClientProxy *self)
+{
+    g_return_val_if_fail (MCD_IS_CLIENT_PROXY (self), FALSE);
+
+    return self->priv->bypass_observers;
+}
+
+gboolean
+_mcd_client_proxy_get_delay_approvers (McdClientProxy *self)
+{
+    g_return_val_if_fail (MCD_IS_CLIENT_PROXY (self), FALSE);
+
+    return self->priv->delay_approvers;
+}
+
 static void
 _mcd_client_proxy_become_incapable (McdClientProxy *self)
 {
@@ -1736,6 +1770,11 @@ _mcd_client_proxy_handle_channels (McdClientProxy *self,
 
         g_hash_table_unref (requests);
 
+        /* Numerical order is correct for all currently supported values:
+         *
+         * (TP_USER_ACTION_TIME_NOT_USER_ACTION == 0) is less than
+         * (normal X11 timestamps, which are 1 to G_MAXUINT32) are less than
+         * (TP_USER_ACTION_TIME_CURRENT_TIME == G_MAXINT64) */
         if (req_time > user_action_time)
             user_action_time = req_time;
 
