@@ -295,7 +295,8 @@ _mcd_connection_attempt (McdConnection *connection)
     if (mcd_account_get_connection_status (connection->priv->account) ==
         TP_CONNECTION_STATUS_DISCONNECTED)
     {
-        _mcd_account_connection_begin (connection->priv->account);
+        /* not user-initiated */
+        _mcd_account_connection_begin (connection->priv->account, FALSE);
     }
     else
     {
@@ -2229,9 +2230,17 @@ static gboolean
 mcd_connection_need_dispatch (McdConnection *connection,
                               const GPtrArray *channels)
 {
+    McdAccount *account = mcd_connection_get_account (connection);
     gboolean any_requested = FALSE, requested_by_us = FALSE;
     guint i;
 
+    if (_mcd_account_needs_dispatch (account))
+    {
+        DEBUG ("Account %s must always be dispatched, bypassing checks",
+               mcd_account_get_object_path (account));
+        return TRUE;
+    }
+
     /* We must _not_ handle channels that have the Requested flag set but that
      * have no McdChannel object associated: these are the channels directly
      * requested to the CM by some other application, and we must ignore them
@@ -2845,7 +2854,10 @@ _mcd_connection_presence_info_is_ready (McdConnection *self)
 
 static void clear_emergency_handles (McdConnectionPrivate *priv)
 {
-  guint n_handles = tp_intset_size (priv->emergency.handles);
+  guint n_handles = 0;
+
+  if (priv->emergency.handles != NULL)
+    n_handles = tp_intset_size (priv->emergency.handles);
 
   /* trawl through the handles and unref them */
   if (n_handles > 0)
