@@ -48,6 +48,63 @@ get_interfaces_quark (void)
     return interfaces_quark;
 }
 
+#define MCD_ACTIVE_OPTIONAL_INTERFACES_QUARK \
+    get_active_optional_interfaces_quark()
+
+static GQuark
+get_active_optional_interfaces_quark (void)
+{
+    static GQuark active_optional_interfaces_quark = 0;
+
+    if (G_UNLIKELY (active_optional_interfaces_quark == 0))
+        active_optional_interfaces_quark =
+            g_quark_from_static_string ("active-optional-interfaces");
+
+    return active_optional_interfaces_quark;
+}
+
+/**
+ * get_active_optional_interfaces:
+ *
+ * DBus Interfaces marked as optional will only be included in the object's
+ * Interfaces property if they appear in this set.
+ *
+ * Returns: a #TpIntset of the active optional interfaces.
+ */
+static TpIntset *
+get_active_optional_interfaces (TpSvcDBusProperties *object)
+{
+    TpIntset *aoi = g_object_get_qdata (G_OBJECT (object),
+                                        MCD_ACTIVE_OPTIONAL_INTERFACES_QUARK);
+
+    if (G_UNLIKELY (aoi == NULL))
+    {
+        aoi = tp_intset_new ();
+
+        g_object_set_qdata_full (G_OBJECT (object),
+                                 MCD_ACTIVE_OPTIONAL_INTERFACES_QUARK,
+                                 aoi,
+                                 (GDestroyNotify) tp_intset_destroy);
+    }
+
+    return aoi;
+}
+
+void
+mcd_dbus_activate_optional_interface (TpSvcDBusProperties *object,
+                                      GType interface)
+{
+    tp_intset_add (get_active_optional_interfaces (object), interface);
+}
+
+gboolean
+mcd_dbus_is_active_optional_interface (TpSvcDBusProperties *object,
+                                       GType interface)
+{
+    return tp_intset_is_member (get_active_optional_interfaces (object),
+                                interface);
+}
+
 static const McdDBusProp *
 get_interface_properties (TpSvcDBusProperties *object, const gchar *interface)
 {
@@ -196,15 +253,6 @@ mcd_dbusprop_get_property (TpSvcDBusProperties *self,
     return TRUE;
 }
 
-static void
-dbusprop_get_cb (TpSvcDBusProperties *self, const GValue *value,
-                 const GError *error, gpointer user_data)
-{
-  DBusGMethodInvocation *context = (DBusGMethodInvocation *) user_data;
-
-  tp_svc_dbus_properties_return_from_get (context, value);
-}
-
 void
 dbusprop_acl_get (TpSvcDBusProperties *self,
                   const gchar *interface,
@@ -232,32 +280,21 @@ dbusprop_get (TpSvcDBusProperties *self,
 {
     GValue value = { 0 };
     GError *error = NULL;
-    const McdDBusProp *property;
 
     DEBUG ("%s, %s", interface_name, property_name);
 
-    /* Look whether the property can support async getting */
-    property = get_mcddbusprop (self, interface_name, property_name, NULL);
-    if (property != NULL && property->async_getprop != NULL)
-    {
-      property->async_getprop (self, property_name, dbusprop_get_cb, context);
-    }
-    else
-    {
-        /* The normal route */
-        mcd_dbusprop_get_property (self, interface_name, property_name,
-                                   &value, &error);
+    mcd_dbusprop_get_property (self, interface_name, property_name,
+                               &value, &error);
 
-        if (error)
-        {
-            dbus_g_method_return_error (context, error);
-            g_error_free (error);
-            return;
-        }
-
-        tp_svc_dbus_properties_return_from_get (context, &value);
-        g_value_unset (&value);
+    if (error)
+    {
+        dbus_g_method_return_error (context, error);
+        g_error_free (error);
+        return;
     }
+
+    tp_svc_dbus_properties_return_from_get (context, &value);
+    g_value_unset (&value);
 }
 
 typedef struct
@@ -269,30 +306,11 @@ typedef struct
 } GetAllData;
 
 static void
-get_all_iter (TpSvcDBusProperties *self,
-              const GValue *value,
-              const GError *error,
-              gpointer user_data)
+get_all_iter (GetAllData *data)
 {
-    GetAllData *data = (GetAllData *) user_data;
-
-    if (self != NULL && value != NULL)
-    {
-        /* This is actually a callback */
-        g_hash_table_insert (data->properties, (gchar *) data->property->name,
-                             tp_g_value_slice_dup (value));
-
-        data->property++;
-    }
-
     if (data->property->name != NULL)
     {
-        if (data->property->async_getprop)
-        {
-          data->property->async_getprop (data->self, data->property->name,
-                                         get_all_iter, data);
-        }
-        else if (data->property->getprop)
+        if (data->property->getprop)
         {
             GValue *out;
 
@@ -304,12 +322,12 @@ get_all_iter (TpSvcDBusProperties *self,
             g_free (out);
 
             data->property++;
-            get_all_iter (NULL, NULL, NULL, data);
+            get_all_iter (data);
         }
         else
         {
             data->property++;
-            get_all_iter (NULL, NULL, NULL, data);
+            get_all_iter (data);
         }
     }
     else
@@ -424,7 +442,7 @@ dbusprop_get_all (TpSvcDBusProperties *self,
 
     data->property = prop_array;
 
-    get_all_iter (NULL, NULL, NULL, data);
+    get_all_iter (data);
 }
 
 void
@@ -477,7 +495,17 @@ mcd_dbus_get_interfaces (TpSvcDBusProperties *self, const gchar *name,
 	if (!iface_data) continue;
 
 	for (id = iface_data; id->get_type; id++)
+        {
+            if (id->optional &&
+                !mcd_dbus_is_active_optional_interface (self,
+                                                        id->get_type ()))
+            {
+                DEBUG ("skipping inactive optional iface %s", id->interface);
+                continue;
+            }
+
 	    g_ptr_array_add (a_ifaces, g_strdup (id->interface));
+        }
     }
     g_ptr_array_add (a_ifaces, NULL);
 
